/*
 *  Calculator -- OpenMoko simple Calculator
 *
 *  Authored by Rodolphe Ortalo <rodolphe.ortalo@free.fr>
 *
 *  Copyright (C) 2007 Rodolphe Ortalo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Public License as published by
 *  the Free Software Foundation; version 2 of the license.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Public License for more details.
 *
 *  Current Version: $Rev$ ($Date$) [$Author$]
 */

#include <stdio.h>

#include <libmokoui/moko-application.h>
#include <libmokoui/moko-finger-tool-box.h>
#include <libmokoui/moko-finger-window.h>
#include <libmokoui/moko-finger-wheel.h>

#include <gtk/gtkalignment.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtk.h>

/*
 * Useful data types
 */
/* state of calculator */
typedef struct {
  double last_operand;
  double current_operand;
  double pre_multiplier;
  double (*func)(double,double);
} calc_state;
/* data passed to signal function via button private data */
typedef struct { guint action; calc_state *state; } button_data;
enum operation {
  notimplemented,
  backspace, clearall, clear,
  zero, one, two, three, four, five, six, seven, eight, nine, ten,
  add, sub, mult, div, equal, minus, point } ;

/*
 * functions for operations
 */
static double noop_func  (double a, double b) { return a; }
static double add_func  (double a, double b) { return a+b; }
static double sub_func  (double a, double b) { return a-b; }
static double mult_func (double a, double b) { return a*b; }
static double div_func  (double a, double b) { return a/b; }

/*
 * Internal variables
 */
static calc_state the_state = { .last_operand = 0.0,
				.current_operand = 0.0,
				.pre_multiplier = 1.0,
				.func = &add_func, };
static GtkWidget* displayed_label;
#define MAX_DISPLAY_CHARS 25
#define DISPLAY_CHARS_FMT "%.15g"
static gchar dispstring[MAX_DISPLAY_CHARS+1] = "1234567890";

#define CALC_ROWS 5
#define CALC_COLS 4
static const gchar *label[CALC_ROWS][CALC_COLS] = {
    {"Bsp", "Clear All", "Clear", "+/-"},
    {"7", "8", "9", "+"},
    {"4", "5", "6", "-"},
    {"1", "2", "3", "*"},
    {"0", ",", "=", "/"},
  };
static enum operation ops[CALC_ROWS][CALC_COLS] = {
    { backspace, clearall, clear, minus},
    { seven, eight, nine, add},
    { four, five, six, sub},
    { one, two, three, mult},
    { zero, point, equal, div},
  };
#if 1
gchar *wnames[CALC_ROWS][CALC_COLS] = {
  {"mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-dialer"},
  {"mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer"},
  {"mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer"},
  {"mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-dialer"},
  {"mokofingerbutton-dialer", "mokofingerbutton-dialer", "mokofingerbutton-black", "mokofingerbutton-dialer"},
};
#else
/* Alternative styling - more flashy */
gchar *wnames[CALC_ROWS][CALC_COLS] = {
  {"mokofingerbutton-black", "mokofingerbutton-black", "mokofingerbutton-black", "mokofingerbutton-orange"},
  {"mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange"},
  {"mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange"},
  {"mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-orange"},
  {"mokofingerbutton-orange", "mokofingerbutton-orange", "mokofingerbutton-black", "mokofingerbutton-orange"},
};
#endif

/*
 * Signal function and static helpers
 */
static void update_state_for_digit(int v)
{
  if (the_state.func == &noop_func) {
    the_state.last_operand = 0.0;
    the_state.func = &add_func;
  }
  if (the_state.pre_multiplier == 1.0 ) {
    the_state.current_operand *= 10.0;
    the_state.current_operand += v;
  } else {
    the_state.current_operand += (v * the_state.pre_multiplier);
    the_state.pre_multiplier /= 10.0;
  }
  snprintf(dispstring,MAX_DISPLAY_CHARS,DISPLAY_CHARS_FMT,the_state.current_operand);
}
static void update_state_for_operation(double (*nextfunc)(double,double))
{
  the_state.last_operand = (*(the_state.func))(the_state.last_operand,the_state.current_operand);
  the_state.current_operand = 0.0;
  the_state.pre_multiplier = 1.0;
  the_state.func = nextfunc;
  snprintf(dispstring,MAX_DISPLAY_CHARS,DISPLAY_CHARS_FMT,the_state.last_operand);
}

void calc_button_pressed( GtkButton* button, enum operation *k)
{
  g_debug( "openmoko-calculator: button pressed" );
  switch (*k) {
  case zero:
    update_state_for_digit(0);
    break;
  case one:
    update_state_for_digit(1);
    break;
  case two:
    update_state_for_digit(2);
    break;
  case three:
    update_state_for_digit(3);
    break;
  case four:
    update_state_for_digit(4);
    break;
  case five:
    update_state_for_digit(5);
    break;
  case six:
    update_state_for_digit(6);
    break;
  case seven:
    update_state_for_digit(7);
    break;
  case eight:
    update_state_for_digit(8);
    break;
  case nine:
    update_state_for_digit(9);
    break;
  case minus:
    the_state.current_operand *= -1;
    snprintf(dispstring,MAX_DISPLAY_CHARS,DISPLAY_CHARS_FMT,the_state.current_operand);
    break;
  case point:
    if (the_state.pre_multiplier == 1.0)
      the_state.pre_multiplier = 0.1;
    break;
  case clearall:
    the_state.last_operand = 0.0;
    the_state.func = &add_func;
  case clear:
    the_state.current_operand = 0.0;
    the_state.pre_multiplier = 1.0;
    snprintf(dispstring,MAX_DISPLAY_CHARS,DISPLAY_CHARS_FMT,the_state.current_operand);
    break;
  case add:
    update_state_for_operation(&add_func);
    break;
  case sub:
    update_state_for_operation(&sub_func);
    break;
  case mult:
    update_state_for_operation(&mult_func);
    break;
  case div:
    update_state_for_operation(&div_func);
    break;
#if 0
  case backspace:
    if (the_state.pre_multiplier == 1.0 ) {
      the_state.current_operand -= (the_state.current_operand % 10.0);
      the_state.current_operand /= 10.0;
    } else {
      the_state.current_operand -= (the_state.current_operand % 10.0);
      the_state.pre_multiplier *= 10.0;
    }
    break;
#endif
  case equal:
    update_state_for_operation(&noop_func); /* FIXME: does not work for renewing a fresh calc */
    break;
  case notimplemented:
    g_debug("openmoko-calculator: operation not-implemented");
    break;
  default:
    g_debug("openmoko-calculator: unknown operation (%i)!", *k);
    break;
  }
  gtk_label_set(displayed_label, dispstring);
}

/*
 * Convenience static function
 */
static GtkTable* calc_panel_init ()
{

  GtkTable *table;
  gint i, j;

  table = GTK_TABLE(gtk_table_new (CALC_ROWS, CALC_COLS, TRUE));
  gtk_table_set_row_spacings(table, 10);
  gtk_table_set_col_spacings(table, 10);

  for (i = 0; i < CALC_ROWS; i++)
    for (j = 0; j < CALC_COLS; j++)
    {
      GtkButton* button = GTK_BUTTON(gtk_button_new_with_label(label[i][j]));

      gtk_table_attach_defaults (table, GTK_WIDGET(button),
                                 j, j + 1,
                                 i, i + 1);
      g_signal_connect (G_OBJECT(button), "clicked", G_CALLBACK(calc_button_pressed),&(ops[i][j]));
      /* TODO: Check if this changes the aspect */
      gtk_widget_set_name(GTK_WIDGET(button), wnames[i][j]);

      gtk_widget_set_size_request (GTK_WIDGET(button), 20, 20);
    }
  return table;
}

/*
 * Command line options definition
 */
static GOptionEntry entries[] = 
{
  /* No options right now except the default ones */
  { NULL }
};


/*
 * Main
 */
int main( int argc, char** argv )
{
    GtkHBox* hbox;
    GtkVBox* vbox;
    GtkTable* table;

    g_debug( "openmoko-calculator starting up" );
    /* Initialize GTK+ */
    gtk_init( &argc, &argv );
    if (argc != 1)
      {
	/* Add init code. */
	GError *error = NULL;
	GOptionContext *context = g_option_context_new ("");
	
	g_option_context_add_main_entries (context, entries, NULL);
	g_option_context_add_group (context, gtk_get_option_group (TRUE));
	g_option_context_parse (context, &argc, &argv, &error);
	
	g_option_context_free (context);
      }

    /* application object */
    /* MokoApplication* app = MOKO_APPLICATION(moko_application_get_instance()); */
    g_set_application_name( "Calculator" );

    /* main window */
    MokoFingerWindow* window = MOKO_FINGER_WINDOW(moko_finger_window_new());

    /* application menu */
    GtkMenu* appmenu = GTK_MENU(gtk_menu_new());
    GtkMenuItem* closeitem = GTK_MENU_ITEM(gtk_menu_item_new_with_label( "Close" ));
    g_signal_connect( G_OBJECT(closeitem), "activate", G_CALLBACK(gtk_main_quit), NULL );
    gtk_menu_shell_append( GTK_MENU_SHELL(appmenu), GTK_WIDGET(closeitem) );
    moko_finger_window_set_application_menu( window, appmenu );

    /* connect close event */
    g_signal_connect( G_OBJECT(window), "delete_event", G_CALLBACK(gtk_main_quit), NULL );

    /* contents */
    hbox = GTK_HBOX(gtk_hbox_new(TRUE, 10));
    vbox = GTK_VBOX(gtk_vbox_new(FALSE, 10));

    displayed_label = gtk_label_new("123");
    gtk_misc_set_alignment(GTK_MISC(displayed_label), 1, 0);
    gtk_label_set_markup(displayed_label, "<span size=\"xx-large\" weight=\"bold\">4321</span>");
    GtkWidget *eventbox1 = gtk_event_box_new ();
    gtk_widget_set_name (eventbox1, "gtkeventbox-black");
    gtk_container_add (GTK_CONTAINER (eventbox1), displayed_label);
    gtk_box_pack_start( GTK_BOX(vbox), eventbox1, FALSE, FALSE, 10);

    table = calc_panel_init();
    gtk_box_pack_start( GTK_BOX(vbox), GTK_WIDGET(table), TRUE, TRUE, 10);

    gtk_box_pack_start( GTK_BOX(hbox), GTK_WIDGET(vbox), TRUE, TRUE, 10);
    moko_finger_window_set_contents(window, GTK_WIDGET(hbox));
    
    /* show everything and run main loop */
    gtk_widget_show_all( GTK_WIDGET(window) );
    g_debug( "calculator entering main loop" );
    gtk_main();
    g_debug( "calculator left main loop" );

    return 0;
}
